/**********************************************************************************

              Fearless Night
    16 Band Stereo Audio Spectrum Analyzer

   Copyright (C) 2009 Brute Force Coding Ltd.
  a division of Really Silly Programming, Inc.

           For use with PCB #9331
 
  2009-03-31 RSP : Created (Version 9331-A)

-----------------------------------------------------------------------------------

  Build:
      - MPLAB 8.30
      -   C30 3.20

  Resources:
      - V.MED, YS LEDS: 10,137 Flash, 15092 RAM w/Full Optimization

-----------------------------------------------------------------------------------

  Features:

      - 2 channels (stereo)
      - 16 frequency bands (each channel)
      - Frequency response is 20 Hz to 20 KHz
      - Frequency resolution is ~ 25 Hz
      - Update rate is ~ 12 frames/sec
      - Adjustable gain setting handles all standard line input levels (0.5Vpp, 1Vpp, 2Vpp)
      - Input level overload/clipping indicator
      - Many different display modes (bar/dot/peak style, smoothing, color, and orientation)
      - Automatic brightness control dims the display in low light

-----------------------------------------------------------------------------------

  Technical Highlights:

      - DMA ADC at 50 KHz (12-bit) sampling rate
      - Optimized 2048 sample 16-bit FFT with saturation detection
      - Net FFT frame rate is ~ 24/second (limited by ADC data acquisition time)
      - Dual 16x16 3-color LED display refreshed at 1KHz (125 Hz net refresh rate)
      - Operates from a standard 5V/1A power cube
      - Front end analog, digital gain, and [dsp]CPU are all Microchip technology

-----------------------------------------------------------------------------------

  Control Switches:

        Mode    = set operating mode (cycles through the display modes)
        Up/Down = set input gain

  Changes are automatically saved in 5 seconds.

  IF (ENABLE_DEMO_DISPLAY compile switch is set)
      Mode held down for 30 secs switches to photography demo
        Mode    = set operating mode (cycles through the display modes)
        Up/Down = brightness control
        (power down to exit demo mode)

**********************************************************************************/

#ifdef __DEBUG
#include "stdio.h"
#include "uart.h"
#endif

#include "sa.h"

// CPU chip configuration at power-on

_FOSCSEL(FNOSC_PRI      // Primary oscillator (XT @ 4/8/10 MHz)
       & IESO_OFF);     //   disable Two-speed Oscillator Startup

_FOSC(FCKSM_CSECMD      // Enable clock switching
    & OSCIOFNC_OFF      //   OSC2 is clock O/P (no other choice when using XT mode oscillator mode)
    & POSCMD_XT);       //   XT oscillator mode (3-10 MHz crystal/resonator)

_FPOR(FPWRT_PWR128);    // Power-on reset timer = 128 msec

_FWDT(FWDTEN_OFF        // Watchdog is enabled/disabled by user software
    & WINDIS_OFF        //   non-window mode
    & WDTPRE_PR128      //   prescaler 1:128 (4 ms timeout)
    & WDTPOST_PS256);   //   postscaler 1:256 (total of ~ 1/2 sec timeout)

_FICD(BKBUG_ON          // Device will Reset in user mode
    & COE_ON            //   Reset in operational mode
    & JTAGEN_OFF        //   JTAG is disabled
    & ICS_PGD1);        //   debug communications on PGC1/EMUC1 and PGD1/EMUD1

_FGS(GSS_OFF            // Code protect Disabled
   & GWRP_OFF);         //   general segment write-protect Disabled

_FBS(RBS_NO_RAM         // No boot RAM
   & BWRP_WRPROTECT_OFF //   boot segment can be written
   & BSS_NO_BOOT_CODE); //   no Boot Segment Program Memory

_FSS(SSS_NO_FLASH       // No Secure Segment
   & SWRP_WRPROTECT_OFF //   secure segment write protect Disabled 
   & RSS_NO_RAM);       //   No Secure Ram

//-----------------------------------------------------------------------------------
// Global Structures
//-----------------------------------------------------------------------------------

// WARNING - DO NOT CHANGE THESE VALUES (the FFT is hard-coded for 2048 points)
#define FFT_BLOCK_LENGTH 2048   // Number of sample points in the FFT
                                // (2X the number of resulting frequency bins)
#define LOG2_BLOCK_LENGTH  11

typedef int   fractional;       // 1.15 format number (signed value, -1 to just under +1)
typedef struct
 { fractional real;
   fractional imag; } 
fractcomplex;

//-----------------------------------------------------------------------------------
// Data tables (constants in program memory)
//-----------------------------------------------------------------------------------

// window function
//    organized as 1/2 of the window (the other half is a mirror of the 1'st half)
extern const int hamming_window_flash[FFT_BLOCK_LENGTH/2]
       __attribute__ ((space(auto_psv)));

// twiddle factors
//    organized as 1024 complex values
extern fractcomplex twiddleAray[FFT_BLOCK_LENGTH/2];
//     __attribute__ ((space(auto_psv)));

// frequency bands (display is the peak value found in each band range)
//    organized as 16 elements of { length } values
extern int freqBin[16]
       __attribute__ ((space(auto_psv)));

//-----------------------------------------------------------------------------------
// Global Data
//-----------------------------------------------------------------------------------

// Bulk Memory Allocation:
// 2K DMA  (1K x 2 bytes input signal)
// 2K Temp (1K x 2 bytes input signal)
// 8K FFT  (2K x 4 bytes complex input signal)
// 2K window constants (could be read directly out of FLASH if RAM becomes scarce)
// ===
// 14K of 16K total in this chip, leaving 2K for global variables and stack

// the FFT processing buffer - accessed both as an int array AND as a fractional array
volatile int   sigAray[ FFT_BLOCK_LENGTH*2 ]
       __attribute__ (( aligned (FFT_BLOCK_LENGTH * 2), far )); // must be aligned for BitReverse to function

// DMA receive buffer
volatile int   dmaBuff[ 1024 ]
       __attribute__ (( space(dma) ));  // entire DMA area is dedicated to receiving FFT data

// temporary storage area
volatile int   tempBuff[ 1024 ]
       __attribute__ (( aligned(2) ));  // stores DMA area (1/2 FFT data) while collecting second half

// windowing function constants, copied from FLASH for faster access from RAM
volatile int   hamming_window_ram[ FFT_BLOCK_LENGTH/2 ];

//-----------------------------------------------------------------------------------

// periodic timer
volatile unsigned int  timer_20msec;    // 1 msec -> 20 msec downcounter

// watchdog interlock
volatile unsigned int  watchdog;        // set by 20msec tick, cleared in foregnd bigloop

//-----------------------------------------------------------------------------------

// DMA transfer control
volatile int  DMA_phase;        // DMA state machine:
                                //  0 = loading 1'st half of DMA buffer [512 words]; takes about 10 msec
                                //  1 = loading 2'nd half of DMA buffer [512 words]; takes about 10 msec
                                //      AT THE SAME TIME, the 1'st half is being copied out to 1'st half of temp buffer
                                //  2 = loading full DMA buffer [1024 words]; takes about 20 msec
                                //      AT THE SAME TIME, the 2'nd half is being copied out to 2'nd half of temp buffer
                                //  3 = complete (signal to foregnd to go ahead and process the data)

volatile int  DMA_xferreq;      // signal to 1msec timer to do a [512 word] transfer
volatile int *DMA_xfersrc;      //   transfer start point (source)
volatile int *DMA_xferdst;      //   transfer start point (destination)
                                // BECAUSE the 1msec interrupt will transfer the 512 word blocks out in < 500 usec,
                                //   the buffers will not overrun each other.

//-----------------------------------------------------------------------------------

// SPI user mutex
volatile int SPI_mode;          // 0 = display, 1 = digital pot

//-----------------------------------------------------------------------------------

// settings are stored in the digital pot EE

volatile unsigned int settings_mode;    // see "derive_settings()" below for mode parameters

volatile unsigned int settings_gain;    // 0..255 (same gain is used for both L/R channels)

volatile unsigned int autosave_lockout; // 20 msec ticks (allows time for EE write to complete)

#define AUTOSAVE_TICKS 5*50;            // 5 seconds @ 20 msec tick rate
volatile unsigned int autosave_mode;    // settings_mode save pending when nonzero (1 EE cell)
volatile unsigned int autosave_gain;    // settings_gain save pending when nonzero (1 EE & 2 volatile cells)

typedef struct tagAUTOSAVE              // autosave SPI write pending for individual EE cells
{   unsigned MODE:1;                    //   user EE cell
    unsigned L_VOLATILE:1;              //   volatile wiper 0 (left channel gain)
    unsigned R_VOLATILE:1;              //   volatile wiper 1 (right channel gain)
    unsigned L_NONVOLATILE:1;           //   nonvolatile wiper 0 (gain setting for both channels)
} AUTOSAVE_STRUCT;
volatile AUTOSAVE_STRUCT autosaveBits;

// derived from "settings_mode"...

#define STYLE_OFF 0
#define STYLE_BAR 1
#define STYLE_DOT 2
volatile unsigned char set_inst_style;
volatile unsigned char set_peak_style;

#define MAX_SMOOTH_STEP 1
volatile unsigned char set_smooth;      // smoothing on/off

volatile unsigned char set_rotate;      // display rotation on/off

//-----------------------------------------------------------------------------------

// input push-buttons

#define KEY_AUTOREPEAT_DELAY 25 // 20 msec ticks before starting autorepeat
#define KEY_AUTOREPEAT_RATE   2 //    autorepeat rate after initial delay

typedef struct tagBUTTON
{   unsigned STATE:8;           // debouncer
    unsigned REPEAT:8;          // auto-repeat timer (applies to up/down buttons only)
} BUTTON_STRUCT;
volatile BUTTON_STRUCT button_mode;
volatile BUTTON_STRUCT button_up;
volatile BUTTON_STRUCT button_down;

//-----------------------------------------------------------------------------------

// demo (photography) mode

volatile unsigned char demo_mode;       // demo mode on/off
volatile unsigned  int demo_timer;      // mode button hold-down timer (1 msec ticks)

//-----------------------------------------------------------------------------------

// signal levels for display
volatile unsigned char sigInst [2][16]; // instantaneous values
volatile unsigned char sigPeak [2][16]; // peak values
volatile unsigned char sigDecay[2][16]; // peak fade timers
#define PEAK_SUSTAIN   25               //   peak sustain time (1/50 sec ticks)
#define PEAK_FADE_RATE 10               //   peak fade rate

// each vertical column is 32 bitz: top 8 red, top 8 grn, bot 8 red, bot 8 grn
volatile unsigned long charset_inst[17];
volatile unsigned long charset_peak[17];
volatile unsigned long charset_mask_bar[17];
volatile unsigned long charset_mask_dot[17];

// display construction area (oriented vertically)
//   this area is copied/converted to disp_buff either as-is (vertical) or rotated 90 degrees
volatile unsigned long disp_temp[32];

// display column strobe control
volatile unsigned int  disp_column;     // column scan index (0..7) that's being strobed (@ 1msec/col)
volatile unsigned char disp_scan[8];    // physical row scan order

// pixels to be sent to the LED array drivers
typedef struct tagDISP_BUFF
{ union
  { struct { unsigned long dw[2*16]; }; // loaded 32 bits at a time
    struct { unsigned int  w [4*16]; }; // clocked out 16 bits at a time
  };
} DISP_BUFF_UNION;
volatile DISP_BUFF_UNION disp_buff;     // column data (sent out in groups of 4 columns)
volatile unsigned int    disp_index;    // column scan index (0..7) that SPI is clocking out now

//-----------------------------------------------------------------------------------
// External subroutines
//-----------------------------------------------------------------------------------

extern fractcomplex* BitReverseComplex( int log2N, fractcomplex* srcCV );
extern unsigned int Window_2048( volatile int *pSrcDst );
extern unsigned int FFT_2048   ( volatile int* pComplexDataArray );
extern unsigned int Complex_Max( int nElements, volatile int** pComplexDataPointer );
extern unsigned int iSqrt      ( unsigned int v );

extern void initialize_disp_scan();

#define COLOR_VU  0
#define COLOR_GRN 1
#define COLOR_YEL 2
#define COLOR_RED 3
extern void initialize_bar_charset( volatile unsigned long *pCharset, unsigned int zColor );
extern void initialize_dot_charset( volatile unsigned long *pCharset, unsigned int zColor );

#define NBR_SETTINGS 13 // count of modes 0..n + default

void derive_settings()
{
    set_smooth = 0;  // no smoothing
    set_rotate = 0;  // no rotation

    switch (settings_mode)
    {
        case 0: // Green inst bar w/ Red peak dot (fast)
            initialize_bar_charset( &charset_inst[0], COLOR_GRN );
            initialize_dot_charset( &charset_peak[0], COLOR_RED );
            set_inst_style = STYLE_BAR;
            set_peak_style = STYLE_DOT;
            break;
        case 1: // Green inst bar w/ Red peak dot (fast), rotated
            initialize_bar_charset( &charset_inst[0], COLOR_GRN );
            initialize_dot_charset( &charset_peak[0], COLOR_RED );
            set_inst_style = STYLE_BAR;
            set_peak_style = STYLE_DOT;
            set_rotate = 1;
            break;
        case 2: // Green inst bar w/ Red peak dot (smoothed)
            initialize_bar_charset( &charset_inst[0], COLOR_GRN );
            initialize_dot_charset( &charset_peak[0], COLOR_RED );
            set_inst_style = STYLE_BAR;
            set_peak_style = STYLE_DOT;
            set_smooth = 1;
            break;
        case 3: // Green inst bar w/ Red peak dot (smoothed), rotated
            initialize_bar_charset( &charset_inst[0], COLOR_GRN );
            initialize_dot_charset( &charset_peak[0], COLOR_RED );
            set_inst_style = STYLE_BAR;
            set_peak_style = STYLE_DOT;
            set_smooth = 1;
            set_rotate = 1;
            break;

        case 4: // Red inst bar w/ Grn peak dot (fast)
            initialize_bar_charset( &charset_inst[0], COLOR_RED );
            initialize_dot_charset( &charset_peak[0], COLOR_GRN );
            set_inst_style = STYLE_BAR;
            set_peak_style = STYLE_DOT;
            break;
        case 5: // Red inst bar w/ Grn peak dot (fast), rotated
            initialize_bar_charset( &charset_inst[0], COLOR_RED );
            initialize_dot_charset( &charset_peak[0], COLOR_GRN );
            set_inst_style = STYLE_BAR;
            set_peak_style = STYLE_DOT;
            set_rotate = 1;
            break;
        case 6: // Red inst bar w/ Grn peak dot (smoothed)
            initialize_bar_charset( &charset_inst[0], COLOR_RED );
            initialize_dot_charset( &charset_peak[0], COLOR_GRN );
            set_inst_style = STYLE_BAR;
            set_peak_style = STYLE_DOT;
            set_smooth = 1;
            break;
        case 7: // Red inst bar w/ Grn peak dot (smoothed), rotated
            initialize_bar_charset( &charset_inst[0], COLOR_RED );
            initialize_dot_charset( &charset_peak[0], COLOR_GRN );
            set_inst_style = STYLE_BAR;
            set_peak_style = STYLE_DOT;
            set_smooth = 1;
            set_rotate = 1;
            break;

        case 8: // YEL inst dot
            initialize_dot_charset( &charset_inst[0], COLOR_YEL );
            set_inst_style = STYLE_DOT;
            set_peak_style = STYLE_OFF;
            set_smooth = 1;
            break;
        case 9: // GRN inst dot
            initialize_dot_charset( &charset_inst[0], COLOR_GRN );
            set_inst_style = STYLE_DOT;
            set_peak_style = STYLE_OFF;
            set_smooth = 1;
            break;
        case 10: // RED inst dot
            initialize_dot_charset( &charset_inst[0], COLOR_RED );
            set_inst_style = STYLE_DOT;
            set_peak_style = STYLE_OFF;
            set_smooth = 1;
            break;

        case 11: // VU peak bar (smoothed)
            initialize_bar_charset( &charset_inst[0], COLOR_VU );
            set_inst_style = STYLE_BAR;
            set_peak_style = STYLE_OFF;
            set_smooth = 1;
            break;
        default: // VU peak bar (smoothed), rotated
            initialize_bar_charset( &charset_inst[0], COLOR_VU );
            set_inst_style = STYLE_BAR;
            set_peak_style = STYLE_OFF;
            set_smooth = 1;
            set_rotate = 1;
    }
}

//-----------------------------------------------------------------------------------
// Main Program Start
//-----------------------------------------------------------------------------------

int main ( void )
{
    int i;  // general purpose variable

//-- configure I/O pins

    _SDI1R = RPIN_SDI;  // RP8  = SDI input

    _RP4R  = RPOR_OC1;  // RP4  = PWM output
    _RP13R = RPOR_SDO1; // RP13 = SDO output
    _RP14R = RPOR_SCK1; // RP14 = SCK output
#ifdef __DEBUG
    _RP11R = RPOR_TXD1; // RP11 = TxD output (debug)
#endif

    CLR_ON;         // all display row drivers OFF
    RCK_OFF;        // display latch signal
    CS_DIGPOT_OFF;  // digital pot [and EE]
    CS_MEM_OFF;     // not used
    LED_DIAG_ON;    // diagnost LED on until initialization is completed
    LED_CLIP_OFF;   // clipping indicator LED off
           // 5432109876543210 
	TRISA = 0b0000000000001111; // RA4 is output; RA1,RA2 are analog in; RA3,RA4 are crystal osc.
    TRISB = 0b1000000101101000;

    AD1PCFGL = 0xFFFF;      // all digital except...
    AD1PCFGLbits.PCFG0 = 0;     // AN0 = left audio
    AD1PCFGLbits.PCFG1 = 0;     // AN1 = right audio
    AD1PCFGLbits.PCFG5 = 0;     // AN5 = photocell

//-- reset handling

	RCONbits.SWDTEN = 0;        // disable watchdog during initialization

    if (RCONbits.TRAPR  != 0) fatal_error_handler(E_TRAP);   // trap conflict occurred ?
    if (RCONbits.IOPUWR != 0) fatal_error_handler(E_OPCODE); // illegal opcode encountered ?
    if (RCONbits.CM     != 0) fatal_error_handler(E_CONFIG); // configuration mismatch occurred ?
    if (RCONbits.WDTO   != 0) fatal_error_handler(E_WDT);    // watchdog timeout occurred ?

//-- krank CPU up on PLL clock

#if CPU_MIPS == 40
#define CPU_N2 0    // N2 = divide-by-2
#define CPU_N2_SET
#endif
#if CPU_MIPS == 20
#define CPU_N2 1    // N2 = divide-by-4
#define CPU_N2_SET
#endif
#if CPU_MIPS == 10
#define CPU_N2 3    // N2 = divide-by-8
#define CPU_N2_SET
#endif
#ifndef CPU_N2_SET
#error "CPU speed is not supported by this code"
#endif

#if XTAL_MHZ == 4
#define CPU_M 80
#define CPU_M_SET
#endif
#if XTAL_MHZ == 8
#define CPU_M 40
#define CPU_M_SET
#endif
#if XTAL_MHZ == 10
#define CPU_M 32
#define CPU_M_SET
#endif
#ifndef CPU_M_SET
#error "XTAL frequency is not supported by this code"
#endif

    // MIPS = fOSC/2; fOSC = fIN*M/(N1*N2)
    PLLFBD             = CPU_M-2;   // FBD = M-2
	CLKDIVbits.PLLPRE  = 2-2;       // N1 = 2; fREF = fIN/N1 = 0.8 to 8 MHz, fVCO = fIN*M/N1 = 100 to 200 MHz
 	CLKDIVbits.PLLPOST = CPU_N2;    // N2 = 2,4,8

    // Clock switch to incorporate PLL
	__builtin_write_OSCCONH(0x03);  // Initiate Clock Switch to Primary Oscillator with PLL (NOSC=0b011)
	__builtin_write_OSCCONL(0x01);  // Start clock switching
	while (OSCCONbits.COSC != 0b011) { /* wait for Clock switch to occur */ } ;
	while (OSCCONbits.LOCK!=1)       { /* wait for PLL to lock */ } ;

//-- diagnostics on COMS @ 9600 baud

#ifdef __DEBUG

    U1MODE = 0x8000;
    U1STA  = 0x0400;

#if CPU_MIPS == 40
#define BAUD_RATE_SET
    U1BRG  = 259;     // w-BRGH=0 (/16)  // 40,000,000 / (16*9600) - 1 = 259
#endif
#if CPU_MIPS == 20
#define BAUD_RATE_SET
    U1BRG  = 129;     // w-BRGH=0 (/16)  // 20,000,000 / (16*9600) - 1 = 129
#endif
#if CPU_MIPS == 10
#define BAUD_RATE_SET
    U1BRG  = 64;      // w-BRGH=0 (/16)  // 10,000,000 / (16*9600) - 1 = 64
#endif
#ifndef BAUD_RATE_SET
#error "CPU speed is not supported by this code"
#endif

extern void Delay_Code_msec( unsigned int msecs );
 Delay_Code_msec(500);

 char sbuf[80];
 sprintf(sbuf,"\r\nBOOT\r\n");
        putsUART1((unsigned int *)sbuf);

#endif

//-- initialize variables

    timer_20msec = 20;

    autosave_mode    = 0;
    autosave_gain    = 0;
    autosave_lockout = 0;

    autosaveBits.MODE          = 0;
    autosaveBits.L_VOLATILE    = 0;
    autosaveBits.R_VOLATILE    = 0;
    autosaveBits.L_NONVOLATILE = 0;

    SPI_mode = 0;

    DMA_xferreq = 0;

    initialize_disp_scan();

    initialize_charset_masks();

    for (i=0; i < FFT_BLOCK_LENGTH/2; i++)
        hamming_window_ram[i] = hamming_window_flash[i];

//-- set interrupt priorities

    // the DMA interrupt must be able to execute well within 1 ADC sample (20 usec),
    //  so configure DMA interrupt to override Timer1 & SPI interrupts in progress

	IPC0bits.T1IP   = 4;    // set Timer1 interrupt priority level to 4
    IPC2bits.SPI1IP = 4;    // set SPI[1] interrupt priority level to 4
	IPC1bits.DMA0IP = 5;    // set DMA CH0 interrupt priority level to 5	

//-- Timer1 provides 1 msec display refresh tick (-> divided by 8 gives 125 Hz net refresh rate)

#define TCYC_1MSEC CPU_MIPS*1000

    TMR1 = 0x0000;
    PR1  = TCYC_1MSEC-1;    // Trigger Timer1 interrupt every 1000 usec (1 KHz)
	IFS0bits.T1IF = 0;		// Clear Timer 1 interrupt
	T1CONbits.TON = 1;		// Start Timer 1

//-- Timer2 clocks the display brightness PWM via "output compare" module

#if CPU_MIPS == 40
#define PWM_BITZ  11
    // PWM freq is ~ 20 KHz
    //   using 1:1 prescale & 11 bit PWM: 1/(2047*(1/40M)*1) = 19.541 KHz
#endif
#if CPU_MIPS == 20
#define PWM_BITZ  11
    // PWM freq is ~ 10 KHz
    //   using 1:1 prescale & 11 bit PWM: 1/(2047*(1/20M)*1) = 9.770 KHz
#endif
#if CPU_MIPS == 10
#define PWM_BITZ  10
    // PWM freq is ~ 10 KHz
    //   using 1:1 prescale & 10 bit PWM: 1/(1023*(1/10M)*1) = 9.775 KHz
#endif
#ifndef PWM_BITZ
#error "CPU speed is not supported by this code"
#endif

    // configure timer 2
    PR2 = (1<<PWM_BITZ)-1;      // OC1RS values 0 to PWM_PR2 result in PWM duty cycles 0 to 100%
    T2CONbits.TCKPS = 0;        // PWM frequency = 1/((PR2+1)*Tcy*prescale)
    T2CONbits.TON   = 1;        // start Timer2
    // configure output compare
    OC1RS = (1<<PWM_BITZ)-1;    // initially off (0%)
    OC1CONbits.OCM = 6;         // PWM mode without fault protection

//-- Timer3 clocks the ADC sample rate

#define TCYC_20USEC CPU_MIPS*20

	PR3  = TCYC_20USEC-1;       // Trigger ADC1 every 20 usec (50 KHz)

//-- SPI setup - communicates with LED display AND digital pot

    // display (74x595) chip clocks serial data in on the rising edge of clock

    // digital pot (using mode 11)
    //   data is clocked in on the rising edge of clock
    //   data is clocked out on the falling edge of clock

    SPI1CON1bits.MSTEN = 1;     // master mode [must set before SMP]
    SPI1CON1bits.CKP   = 1;     // clock idle is high, active is low
    SPI1CON1bits.CKE   = 0;     // peripheral clocks in data on rising edge, state changes on falling edge
    SPI1CON1bits.SMP   = 0;     // peripheral data is sampled on rising edge, state changes on falling edge
#if CPU_MIPS == 40
#define SPI_SPEED_SET
    SPI1CON1bits.PPRE  = 2;     // 4:1 primary prescale
    SPI1CON1bits.SPRE  = 4;     // 4:1 secondary prescale (results in a 2.5 MHz SPI clock)
#endif
#if CPU_MIPS == 20
#define SPI_SPEED_SET
    SPI1CON1bits.PPRE  = 2;     // 4:1 primary prescale
    SPI1CON1bits.SPRE  = 6;     // 2:1 secondary prescale (results in a 2.5 MHz SPI clock)
#endif
#if CPU_MIPS == 10
#define SPI_SPEED_SET
    SPI1CON1bits.PPRE  = 3;     // 1:1 primary prescale
    SPI1CON1bits.SPRE  = 4;     // 4:1 secondary prescale (results in a 2.5 MHz SPI clock)
#endif
#ifndef SPI_SPEED_SET
#error "CPU speed is not supported by this code"
#endif
    SPI1STATbits.SPIROV= 0;     // clear receive overflow
  //SPI1STATbits.SPIEN = 1;     // enable SPI (configures I/O pins)

    // set SPI CLK digital output pin to idle state (high),
    //   because it's necessary to turn off the SPI module (pin reverts to digital output)
    //   in order to change the MODE16 setting.
    LATBbits.LATB14 = 1;

//-- ADC setup

	AD1CON1bits.FORM   = 1;		// Data Output Format: Signed Integer
	AD1CON1bits.SSRC   = 2;		// Sample Clock Source: GP Timer(3) starts conversion
	AD1CON1bits.ASAM   = 1;		// ADC Sample Control: Sampling begins immediately after conversion
    AD1CON1bits.AD12B  = 1;	    // 12-bit ADC operation
	AD1CON1bits.ADDMABM= 1; 	// DMA buffers build mode: order of conversion

    // No channel scan for CH0+, Use MUX A, SMPI = 1 per interrupt, Vref = AVdd/AVss
	AD1CON2 = 0x0000;

    // Set Samples and bit conversion time
	AD1CON3bits.ADRC = 0;		// ADC Clock is derived from System Clock

#if CPU_MIPS == 40
#define ADC_SPEED_SET
	AD1CON3bits.ADCS = 45;		// ADC sampling time (12 bit) is:    3*Tad
								// ADC Conversion Time (12-bit) is: 14*Tad
                                // ADC sample+conversion time is:   17*Tad, which must be < 20 usec
                                //                                  so Tad must be < 1.176 usec
                                //              ADCS = (Tad/Tcy)-1, so ADCS must be < 46.05
                                //     choosing ADCS = 45, Tad = (46*1/40M) = 1.150 usec
                                // ADC sample+conversion time is:   17*Tad = 19.550 usec
#endif
#if CPU_MIPS == 20
#define ADC_SPEED_SET
	AD1CON3bits.ADCS = 22;		// ADC sampling time (12 bit) is:    3*Tad
								// ADC Conversion Time (12-bit) is: 14*Tad
                                // ADC sample+conversion time is:   17*Tad, which must be < 20 usec
                                //                                  so Tad must be < 1.176 usec
                                //              ADCS = (Tad/Tcy)-1, so ADCS must be < 22.52
                                //     choosing ADCS = 22, Tad = (23*1/20M) = 1.150 usec
                                // ADC sample+conversion time is:   17*Tad = 19.550 usec
#endif
#if CPU_MIPS == 10
#define ADC_SPEED_SET
	AD1CON3bits.ADCS = 10;		// ADC sampling time (12 bit) is:    3*Tad
								// ADC Conversion Time (12-bit) is: 14*Tad
                                // ADC sample+conversion time is:   17*Tad, which must be < 20 usec
                                //                                  so Tad must be < 1.176 usec
                                //              ADCS = (Tad/Tcy)-1, so ADCS must be < 10.76
                                //     choosing ADCS = 10, Tad = (11*1/10M) = 1.100 usec
                                // ADC sample+conversion time is:   17*Tad = 18.700 usec
#endif
#ifndef ADC_SPEED_SET
#error "CPU speed is not supported by this code"
#endif

    // it takes 20 usec for the ADC to powerup, so it's left on all the time
	AD1CON1bits.ADON = 1;		// Turn on the A/D converter

//-- DMA setup - Channel 0 does the audio signal data transfer (ADC)

	DMA0CONbits.AMODE = 0;		// Configure DMA for register indirect mode with post-increment
	DMA0CONbits.MODE  = 0;		// Configure DMA for continuous mode
    DMA0CONbits.SIZE  = 0;      // word transfers
	DMA0PAD= (int)&ADC1BUF0;    // peripheral data address (must be word aligned for ADC!)
	DMA0REQ = 13;				// Select ADC1 as DMA Request source
	IFS0bits.DMA0IF  = 0;		// Clear the DMA interrupt flag bit

//-- load settings from EE [in the digital pot chip]

  //SPI1STATbits.SPIEN = 0;     // must turn off SPI module to change 8/16 mode
    SPI1CON1bits.MODE16 = 1;    // all digital pot transfers are 16-bit
    SPI1STATbits.SPIEN = 1;

    // using polled SPI, so INTERRUPTS MUST BE DISABLED AT THIS POINT

    // read mode setting (stored in user EE area)
    CS_DIGPOT_ON
    SPI1BUF = POT_CMD_READ | POT_EE_USER; //  user EE data [9 bits wide]
      while (SPI1STATbits.SPIRBF == 0) { /* wait for SPI receive to complete */ } ;
      settings_mode = SPI1BUF;
      settings_mode &= 0xFF;    // mask off to data only
      if (settings_mode >= NBR_SETTINGS) settings_mode = 0; // protect against bad data
    CS_DIGPOT_OFF

    // digital pot chip requires 50 nsec min to next /CS
    derive_settings();

    // read gain setting (stored in nonvolatile wiper 0)
    CS_DIGPOT_ON
    SPI1BUF = POT_CMD_READ | POT_EE_WIPER;
      while (SPI1STATbits.SPIRBF == 0) { /* wait for SPI receive to complete */ } ;
      settings_gain = SPI1BUF;
      settings_gain &= 0xFF;    // mask off to data only
    CS_DIGPOT_OFF

    // override right channel to match left channel gain
    autosaveBits.R_VOLATILE = 1;

//-- initialization complete

    CLR_OFF;                    // enable display column driver
    LED_DIAG_OFF;               // diagnostic lED off

//-- enable interrupts

    IEC0bits.DMA0IE = 1;		// Enable DMA interrupt
	IEC0bits.T1IE   = 1;        // Enable Timer 1 interrupt

//-- enable Watch Dog Timer

    ClrWdt();
#ifndef __DEBUG // no watchdog when using ICD2 for debugging
    RCONbits.SWDTEN = 1;
#endif

//-----------------------------------------------------------------------------------
// BIGLOOP
//-----------------------------------------------------------------------------------

  while (1)
  {

//  {__asm__ volatile ("btg LATB,#2");}  // frame rate test (measure freq on diag LED)

    // watchdog will fire if either the foreground bigloop crashes or the periodic timer fails

    if (watchdog != 0) { ClrWdt();  watchdog = 0; }

#ifdef ENABLE_DEMO_DISPLAY
  if (demo_mode == 0)
  {
#endif

    // wait for data acquisition to complete
    //
    // @ 40 MIPS, about 90% of the available CPU time is spent here idling
    // @ 10 MIPS, a little more than 50% of the available CPU time is spent here idling

//  {__asm__ volatile ("bset LATB,#2");} // idle CPU timing test (measure duty cycle on diag LED)

    while (DMA0CONbits.CHEN != 0) { /* wait for DMA idle */ };

//  {__asm__ volatile ("bclr LATB,#2");} // idle CPU timing test (measure duty cycle on diag LED)

    // transfer input signal to FFT processing buffer (only 1'st half [4Kb] is used initially)

    asm volatile( " mov #_tempBuff,w0 \n"           // tempBuff[0]
                  " mov #_sigAray,w1 \n"            // sigAray [0]
                  " repeat #(1024-1) \n"            // xfer 1024 x 2 bytes == 2Kb == 1Kw
                  " mov [w0++],[w1++]\n"
        :               // no outputs 
        :               // no inputs      
        : "w0", "w1" ); // clobbers w0,w1

    asm volatile( " mov #_dmaBuff,w0 \n"            // dmaBuff[0]
                  " mov #(_sigAray+2*1024),w1 \n"   // sigAray[1024]
                  " repeat #(1024-1) \n"            // xfer 1024 x 2 bytes == 2Kb == 1Kw
                  " mov [w0++],[w1++]\n"
        :               // no outputs 
        :               // no inputs      
        : "w0", "w1" ); // clobbers w0,w1

    // do signal acquisition on other channel while doing analysis on this channel

    int this_ch;

    if (AD1CHS0 == ADCH_L)
       {
         // switch to other audio channel
         AD1CHS0 = ADCH_R;  this_ch = 0; 
       }
    else
       { // BUT FIRST, read the photocell sensor
         AD1CHS0 = ADCH_PHOT;
         AD1CON1bits.FORM   = 0;    // Data Output Format: Integer
         AD1CON1bits.ASAM   = 0;    // sampling begins when SAMP is set
         AD1CON1bits.SSRC   = 7;	// Sample Clock Source: Auto-convert
         AD1CON3bits.SAMC   = 3;    // auto sample time = 3*Tad
         // stall until ADC conversion is complete
         AD1CON1bits.DONE   = 0;    // clear conversion complete indicator
         AD1CON1bits.SAMP   = 1;    // start ADC conversion (do sampling, followed by conversion)
         while (AD1CON1bits.DONE == 0) { /* wait for ADC conversion to complete */ } ;
         i = ADC1BUF0;              // read ambient light level
         // set brightness level
// the LEDs will dim ~8-15% after 1000 hours, so they are set a little bright initially
#if LED_TYPE == LED_TYPE_KB
#define LED_BRITE_SET
// prototype #1...
#define MIN_BRITE    5  // %
#define MAX_BRITE   75  // %
         i = i >> (12-PWM_BITZ);    // downshift from 12 bit ADC to nbr bits in PWM
#endif

#if LED_TYPE == LED_TYPE_YS
#define LED_BRITE_SET
// prototype #2...
#define MIN_BRITE   25  // %
#define MAX_BRITE  100  // %
#endif

#ifndef LED_BRITE_SET
#error "LED array type is not supported by this code"
#endif
#define MIN_PWM ((unsigned int) (((1<<PWM_BITZ)-1) * (MIN_BRITE/100.)))
#define MAX_PWM ((unsigned int) (((1<<PWM_BITZ)-1) * (MAX_BRITE/100.)))
         if (i < MIN_PWM) i = MIN_PWM;  // minimum brightness rail (display is never fully off)
         if (i > MAX_PWM) i = MAX_PWM;  // maximum brightness rail
         OC1RS = i ^ ((1<<PWM_BITZ)-1); // negative logic on PWM output
         // restore for audio signal sampling
         AD1CON1bits.FORM   = 1;	// Data Output Format: Signed Integer
         AD1CON1bits.SAMP   = 1;    // begin sampling for 1'st DMA ADC conversion
         AD1CON1bits.SSRC   = 2;	// Sample Clock Source: GP Timer(3) starts conversion
         AD1CON1bits.ASAM   = 1;    // ADC Sample Control: Sampling begins immediately after conversion
         AD1CON3bits.SAMC   = 0;

         // ok, now ready to switch to the other audio channel
         AD1CHS0 = ADCH_L;  this_ch = 1; 
       }

    DMA_phase   = 0;            // initial transfer is to 1'st 1/2 of DMA buffer
    DMA_xferreq = 0;
    DMA_xfersrc = &dmaBuff [0]; //   then copyout 1'st 1/2 of DMA buffer to temp buffer
    DMA_xferdst = &tempBuff[0];

    DMA0STA = 0;                // beginning of DMA buffer
    DMA0CNT = (512)-1;			// fill 1/2 buffer (512 words == 1Kb)

    IFS0bits.AD1IF   = 0;		// Clear the A/D interrupt flag
    DMA0CONbits.CHEN = 1;       // start DMA transfer
	TMR3 = 0x0000;
    T3CONbits.TON    = 1;       // Start Timer 3

    // perform FFT analysis

    unsigned int is_saturated;  // saturation indicator

    Window_2048( &sigAray[0] ); // apply hamming window to reduce spectral leakage

	BitReverseComplex( LOG2_BLOCK_LENGTH-1, (fractcomplex *)&sigAray[0] );

	is_saturated = FFT_2048( &sigAray[0] );

    // indicate if clipping (saturation) is occurring

    if (is_saturated != 0 && settings_gain != 0)
    {  LED_CLIP_ON
#ifdef __DEBUG
        sprintf(sbuf,"\r\nSATURATION CHAN(%u) CLIP(%X)\r\n",this_ch,is_saturated);
        putsUART1((unsigned int *)sbuf);
#endif
    } else LED_CLIP_OFF

    // extract FFT results (find max in each frequency bin range)

    {   volatile int *p;
    
        for (i=0, p = sigAray+2; i < 16; i++)
        {
            unsigned int  u;
    
            // find max value in bin
            //   max value will be +32K (15 bits) max
            //   we will consider +16K (14 bits) to be full scale
            u = Complex_Max( freqBin[i], &p ); // autoincrements p
    
            // convert value to power
            u = iSqrt(u);
    
            // convert power to display charset value (0..127 -> 0..15)
            //    power value full scale is sqrt(16K) = 127
            if ((u & 0x02) != 0) u += 0x04; // round
            if (u > 63)
                u = 16;     // 16 = overrange
            else
                u = u >> 2; // 0..15
    
            // force baseline setting
            if (u == 0) u = 1;
    
            // handle smoothing setting
            if (set_smooth != 0)
            {
                int du;  du = u - sigInst[this_ch][i];
    
                if (du > 0 && du <  2) du =  2; // min change (if any) is 1 step
                if (du < 0 && du > -2) du = -2; // min change (if any) is 1 step
     
                u = sigInst[this_ch][i] + du/2;
            }
    
            // store instantaneous display value
            sigInst[this_ch][i] = u;
    
            // store peak display value
            if (u >= sigPeak[this_ch][i])
            {
                sigPeak [this_ch][i] = u;
                sigDecay[this_ch][i] = PEAK_SUSTAIN;
            }
        }
    }

#ifdef __DEBUG
 { int j;
 sprintf(sbuf,"\r\n%u:",this_ch);
        putsUART1((unsigned int *)sbuf);
 for (j=0; j < 16; j++) {
   sprintf(sbuf,"  %u",sigInst[this_ch][j]);
        putsUART1((unsigned int *)sbuf); }
 }
#endif // DEBUG

#ifdef ENABLE_DEMO_DISPLAY
  } // if (!demo_mode)
  else
  {
    //-- static/stable display for photography

    // left channel
    sigInst[1][ 0] =   sigPeak[1][ 0] =  1;
    sigInst[1][ 1] =   sigPeak[1][ 1] =  2;
    sigInst[1][ 2] =   sigPeak[1][ 2] =  9;
    sigInst[1][ 3] =   sigPeak[1][ 3] =  5;
    sigInst[1][ 4] =   sigPeak[1][ 4] = 13;
    sigInst[1][ 5] =   sigPeak[1][ 5] = 16;
    sigInst[1][ 6] =   sigPeak[1][ 6] =  8;
    sigInst[1][ 7] =   sigPeak[1][ 7] =  5;
    sigInst[1][ 8] =   sigPeak[1][ 8] =  2;
    sigInst[1][ 9] =   sigPeak[1][ 9] =  1;
    sigInst[1][10] =   sigPeak[1][10] =  2;
    sigInst[1][11] =   sigPeak[1][11] =  7;
    sigInst[1][12] =   sigPeak[1][12] =  8;
    sigInst[1][13] =   sigPeak[1][13] =  3;
    sigInst[1][14] =   sigPeak[1][14] =  2;
    sigInst[1][15] =   sigPeak[1][15] =  1;
    // right channel
    sigInst[0][ 0] =   sigPeak[0][ 0] =  1;
    sigInst[0][ 1] =   sigPeak[0][ 1] =  3;
    sigInst[0][ 2] =   sigPeak[0][ 2] = 10;
    sigInst[0][ 3] =   sigPeak[0][ 3] =  4;
    sigInst[0][ 4] =   sigPeak[0][ 4] = 14;
    sigInst[0][ 5] =   sigPeak[0][ 5] = 15;
    sigInst[0][ 6] =   sigPeak[0][ 6] =  8;
    sigInst[0][ 7] =   sigPeak[0][ 7] =  5;
    sigInst[0][ 8] =   sigPeak[0][ 8] =  2;
    sigInst[0][ 9] =   sigPeak[0][ 9] =  1;
    sigInst[0][10] =   sigPeak[0][10] =  2;
    sigInst[0][11] =   sigPeak[0][11] =  7;
    sigInst[0][12] =   sigPeak[0][12] =  7;
    sigInst[0][13] =   sigPeak[0][13] =  3;
    sigInst[0][14] =   sigPeak[0][14] =  2;
    sigInst[0][15] =   sigPeak[0][15] =  1;
/*
    // constant brightness setting (full on == 1024)
    //      250 is good for prototype #1 (bright LEDs)
    OC1RS = 500 ^ ((1<<PWM_BITZ)-1); // negative logic on PWM output
*/
  } // if (demo_mode)
#endif

    //-- build [temporary] display buffer from signal data

    {   volatile unsigned long *pL;  pL = &disp_temp[ 0];
        volatile unsigned long *pR;  pR = &disp_temp[16];
        unsigned char v, pk;

        switch (set_inst_style)
        {
            case STYLE_OFF: // style peak_off is invalid for this case

                for (i=0; i < 16; i++, pR++, pL++)
                {
                    *pR = charset_peak[sigPeak[0][i]];
                    *pL = charset_peak[sigPeak[1][i]];
                }
                break;

            case STYLE_BAR:

                if (set_peak_style != 0)

                    for (i=0; i < 16; i++, pR++, pL++)
                    {
                        // limit inst to peak-1
                        v  = sigInst[0][i];  pk = sigPeak[0][i];  if (v >= pk && pk > 0) v = pk-1;
                        // load peak & mask off inst from peak
                        *pR = (charset_peak[pk] & charset_mask_bar[v]) | charset_inst[v];
    
                        // limit inst to peak-1
                        v  = sigInst[1][i];  pk = sigPeak[1][i];  if (v >= pk && pk > 0) v = pk-1;
                        // load peak & mask off inst from peak
                        *pL = (charset_peak[pk] & charset_mask_bar[v]) | charset_inst[v];
                    }

                else // inst bar ony

                    for (i=0; i < 16; i++, pR++, pL++)
                    {
                        *pR = charset_inst[sigInst[0][i]];
                        *pL = charset_inst[sigInst[1][i]];
                    }

                break;

            case STYLE_DOT:
        
                switch (set_peak_style)
                {
                    case STYLE_DOT: // inst dot + peak dot: peak "overwrites" inst dot if they're the same

                        for (i=0; i < 16; i++, pR++, pL++)
                        {
                            // load peak
                            pk = sigPeak[0][i];  *pR = charset_peak[ pk ];
                            // load inst
                            v  = sigInst[0][i];  if (pk != v) *pR |= charset_inst[v];
    
                            // load peak
                            pk = sigPeak[1][i];  *pL = charset_peak[ pk ];
                            // load inst
                            v  = sigInst[1][i];  if (pk != v) *pL |= charset_inst[v];
                        }
                        break;

/* this mode (peak bar) sucks so much that it isn't used...    
                    case STYLE_BAR: // inst dot + peak bar: inst dot "overwrites" peak bar

                        for (i=0; i < 16; i++, pR++, pL++)
                        {
                            v = sigInst[0][i];  pk = sigPeak[0][i];
                            *pR = (charset_peak[pk] & charset_mask_dot[v]) | charset_inst[v];

                            v = sigInst[1][i];  pk = sigPeak[1][i];
                            *pL = (charset_peak[pk] & charset_mask_dot[v]) | charset_inst[v];
                        }
                        break;
*/    
                    default: // inst dot only

                        for (i=0; i < 16; i++, pR++, pL++)
                        {
                            *pR = charset_inst[sigInst[0][i]];
                            *pL = charset_inst[sigInst[1][i]];
                        }
                }
        }
    }

    // copy/convert temporary display buffer to actual display buffer
    //
    //   disp_temp contains (DWORDS) which are the columns are in display order (L->R), vertical orientation
    //    MSB gets clocked out first, LSB last

    if (set_rotate == 0) // normal vs. rotated graph display

        for (i=0; i < 32; i++) disp_buff.dw[i] = disp_temp[i];

    else
    {   // arbitrarily map 32x32(=1024) bits to rotate the graph images 90 degrees
        //
        //   actual implementation: ~ 3328 cycles = 333 usec
        //
        // don't hurt yourself looking at the code; it's a generated file

#if PCB_TYPE == PCB_SMALL
#define LED_ROTATOR_SET
        #include "SA_rotator_s.c"
#endif

#if PCB_TYPE == PCB_MED
#define LED_ROTATOR_SET
        #include "SA_rotator_m.c"
#endif

#ifndef LED_ROTATOR_SET
//#error "Display rotation is not supported by this code"
#endif
    }

  } // while (1)
}


/*=============================================================================
_DMA0Interrupt(): ISR name is chosen from the device linker script.
=============================================================================*/

/*
    Sample rate = 50 KHz (= 20 usec)
       for 2048 samples -> 40.96 msec = ~24/sec max possible rate (mono)
    target update rate is > 10/sec (stereo)
       actual achieved rate is ~12/sec
*/

void __attribute__((interrupt, no_auto_psv)) _DMA0Interrupt(void)
{
	IFS0bits.DMA0IF  = 0;       // clear interrupt source

    DMA0CONbits.CHEN = 0;       // reset DMA transfer (simulate one-shot mode in continuous mode)

    if (DMA_phase == 0)      // 1'st half of DMA buffer has been loaded
    {
        // start next block load
        DMA0STA = 512*2;        // start @ halfway point in DMA buffer
       	DMA0CONbits.CHEN = 1;   // start DMA xfer; timer3 is still running

        // perform loadout of this block in 1msec tick handler
        DMA_xferreq = 1;
    }
    else if (DMA_phase == 1) // 2'nd half of DMA buffer has been loaded
    {
        // start final block load
        DMA0STA = 0;            // back to beginning of buffer
        DMA0CNT = (1024)-1;     // load full DMA buffer this time
       	DMA0CONbits.CHEN = 1;   // start DMA xfer; timer3 is still running

        // perform loadout of this block in 1msec tick handler
        DMA_xferreq = 1;
    }
    else                     // full DMA buffer has been loaded
    {
    	T3CONbits.TON = 0;		// stop Timer 3

        // tempBuff + dmaBuff now contains 2048 sample points [2Kw == 4Kb]
    }

    DMA_phase++;
}


/*=============================================================================
_T1Interrupt(): ISR name is chosen from the device linker script.
=============================================================================*/

/*
    Refresh tick = 1 msec (results in 125 Hz display refresh rate)
*/

void __attribute__((interrupt, no_auto_psv)) _T1Interrupt(void)
{
	IFS0bits.T1IF = 0;		// Clear interrupt source

//-- send the next display column to the display SR chain

    // The display SR is organized as:
    //   4 groups of columns, 16 bits high each, last[rightmost] display group first
    //   followed by the column drive byte
    //   ALL OF WHICH have scrambled bit arrangement due to PCB layout optimization

    disp_column = (disp_column+1) & 0x07;

    // trap here if SPI interrupt is already enabled
    if (IEC0bits.SPI1IE != 0) fatal_error_handler(E_SPI);   // program logic error (SPI should be available!)
    // trap here if digital pot transfer is being interrupted
    if (SPI_mode != 0)        fatal_error_handler(E_MUTEX); // program logic error (pot CS should be deasserted!)

    // initiate SPI xfer using interrupts (DMA is fully allocated to signal data transfers)
    disp_index = 0;
    SPI1STATbits.SPIEN = 0;     // must turn off SPI module to change 8/16 mode
    SPI1CON1bits.MODE16 = 1;    // first 8 transfers will be 16-bit
    SPI1STATbits.SPIEN = 1;
    SPI1STATbits.SPIROV = 0;
    // start 1'st transfer manually, the remainder will be interrupt driven
    SPI1BUF = disp_buff.w[2*24+(disp_column<<1)];           // last column must be sent out first
    IFS0bits.SPI1IF = 0;
    IEC0bits.SPI1IE = 1;        // enable SPI interrupt

//-- copy DMA block to temp storage when requested

    if (DMA_xferreq != 0)
    {
        DMA_xferreq = 0;

        // the DMA buffer will overrun in about 20usec * 512 = 10.240 msec
        //   worst case latency to get here is a little over 1 msec
        //   @ 40 MIPS, moving 512 words will take about 25 usec
        //   @ 10 MIPS, moving 512 words will take about 100 usec
        // so there should be plenty of time to move the DMA data out of the way

        asm volatile( " mov _DMA_xfersrc,w0 \n"
                      " mov _DMA_xferdst,w1 \n"
                      " repeat #(512-1) \n"
                      " mov [w0++],[w1++]\n"
            :               // no outputs 
            :               // no inputs      
            : "w0", "w1" ); // clobbers w0,w1

        // prepare to move 2'nd block next
        DMA_xfersrc = &dmaBuff [512];
        DMA_xferdst = &tempBuff[512];
    }

//-- human interface tick (20 msec)

    if (--timer_20msec == 0)
    {
        int i, ch;

        //-- perform peak signal decay timing

        if (demo_mode == 0)
        for (ch=0; ch < 2; ch++)
            for (i=0; i < 16; i++)
                if (sigDecay[ch][i] > 0)
                    if (--sigDecay[ch][i] == 0)
                        if (--sigPeak[ch][i] > 0)
                             sigDecay[ch][i] = PEAK_FADE_RATE;

        //-- perform pushbutton debouncing
    
        if (BTN_MODE != 0) // button is released
            button_mode.STATE = 0;
        else               // button is pressed
            if (button_mode.STATE == 0)
                button_mode.STATE = 1;  // debounce for 20 msec
            else if (button_mode.STATE == 1)
            {
                button_mode.STATE = 2;  // one-shot button
                demo_timer = 50*30;     // 20 tick/msec => 30 seconds

                // handle MODE button
                settings_mode++;  if (settings_mode >= NBR_SETTINGS) settings_mode = 0;
                autosave_mode = AUTOSAVE_TICKS; // update EE in 5 seconds if no further changes
                derive_settings();
            }
#ifdef ENABLE_DEMO_DISPLAY
            else // state 2 (button is being held down)
            {
                if (demo_timer == 0) demo_mode = 1;  else demo_timer -= 1;
            }
#endif

        i = 0;
        if (BTN_UP != 0)   // button is released
            button_up.STATE = 0;
        else               // button is pressed
            if (button_up.STATE == 0)
                button_up.STATE = 1;  // debounce for 20 msec
            else if (button_up.STATE == 1)
            {
                button_up.STATE = 2;  // auto-repeat button
                button_up.REPEAT = KEY_AUTOREPEAT_DELAY;
                i = 1;
            }
            else           // button is held
                if (--button_up.REPEAT == 0)
                {
                    button_up.REPEAT = KEY_AUTOREPEAT_RATE;
                    i = 1;
                }
        if (i)
          if (demo_mode == 0)
          {
            // handle UP button
            if (settings_gain < 255)
            {
                settings_gain++;
                autosaveBits.L_VOLATILE = 1;    // update gain setting immediately (in < 5 msec)
                autosaveBits.R_VOLATILE = 1;
                autosave_gain = AUTOSAVE_TICKS; // update EE in 5 seconds if no further changes
            }
          }
          else // up/down controls demo brightness
            OC1RS--;

        i = 0;
        if (BTN_DOWN != 0) // button is released
            button_down.STATE = 0;
        else               // button is pressed
            if (button_down.STATE == 0)
                button_down.STATE = 1;  // debounce for 20 msec
            else if (button_down.STATE == 1)
            {
                button_down.STATE = 2;  // auto-repeat button
                button_down.REPEAT = KEY_AUTOREPEAT_DELAY;
                i = 1;
            }
            else           // button is held
                if (--button_down.REPEAT == 0)
                {
                    button_down.REPEAT = KEY_AUTOREPEAT_RATE;
                    i = 1;
                }
        if (i)
          if (demo_mode == 0)
          {
            // handle down button
            if (settings_gain > 0)
            {
                settings_gain--;
                autosaveBits.L_VOLATILE = 1;    // update gain setting immediately (in < 5 msec)
                autosaveBits.R_VOLATILE = 1;
                autosave_gain = AUTOSAVE_TICKS; // update EE in 5 seconds if no further changes
            }
          }
          else // up/down controls demo brightness
            OC1RS++;

        //-- perform autosave timing (save setting to EE after 5 seconds of no change)

        if (autosave_mode != 0)     // write mode to 1'st user EE cell
            if (--autosave_mode == 0)
                autosaveBits.MODE = 1;

        if (autosave_gain != 0)     // write gain value to left channel only
            if (--autosave_gain == 0)
                autosaveBits.L_NONVOLATILE = 1;

        //-- perform EE write timing (digital pot chip needs 10 msec max to complete an EE write)

        if (autosave_lockout != 0)  // provides 20 to 40 msec delay between EE writes
            autosave_lockout--;

        //-- end of 20 msec tick
    
        timer_20msec = 20;

        watchdog = 1;               // pet the watchdog in foreground bigloop
    }
}


/*=============================================================================
_SPI1Interrupt(): ISR name is chosen from the device linker script.
=============================================================================*/

/*
    SPI [transfer complete] interrupt
*/

void __attribute__((interrupt, no_auto_psv)) _SPI1Interrupt(void)
{
    IFS0bits.SPI1IF = 0;        // clear interrupt source

    // must discard received data, even though SPI is being used for transmit only
    //   (receive overflow will lock up the SPI interrupt)
    SPI1STATbits.SPIROV = 0;    // clear overrun error (caused by not reading SPI1BUF)

    // SPI could be sending data to the display OR the digital pot

    if (SPI_mode != 0)
    {
        IEC0bits.SPI1IE = 0;    // end (digital pot) SPI interrupt
        // min 100 nsec after SPI completes to de-asserting CS
        CS_DIGPOT_OFF;          // de-assert CS to digital pot chip
        // release SPI mutex
        SPI_mode = 0;           // digital pot SPI write is complete
    }
    else
    {
        // send next display data (if any)
        //
        //    @ 2.5 MHz SPI clock, sending 17 bytes will take < 100 usec
    
        disp_index++;
    
        if (disp_index < 8)       // sending LED column data [words]
        {
            SPI1BUF = disp_buff.w[ (disp_column << 1) + (disp_index & 1) + (24*2-((disp_index >> 1) << 4)) ];
        }
        else if (disp_index == 8) // sending column drive select byte
        {
            SPI1STATbits.SPIEN = 0;     // must turn off SPI module to change 8/16 mode
            SPI1CON1bits.MODE16 = 0;    // last transfer is 8-bit
            SPI1STATbits.SPIEN = 1;
            SPI1BUF = disp_scan[ disp_column ];
        }
        else                      // last SPI transfer is complete
        {
            // the minimum RCLK pulse width is 5 nsec for 74VHC595 chip
            //   HOWEVER, it must pass through the 74HCT365 chip first
            //   AND THEN it must drive a fan-out of 16 74VHC595 chips
            RCK_ON;
#if CPU_MIPS == 40
#define RCK_SPEED_SET
              Nop();    //  25 ns
              Nop();    //  50 ns
              Nop();    //  75 ns
              Nop();    // 100 ns
              Nop();    // 125 ns
              Nop();    // 150 ns
              Nop();    // 175 ns
              Nop();    // 200 ns
              Nop();    // 225 ns
              Nop();    // 250 ns
              Nop();    // 275 ns
              Nop();    // 300 ns
              Nop();    // 325 ns
              Nop();    // 350 ns
              Nop();    // 375 ns
              Nop();    // 400 ns [2.5 MHz]
#endif
#if CPU_MIPS == 20
#define RCK_SPEED_SET
              Nop();    //  50 ns
              Nop();    // 100 ns
              Nop();    // 150 ns
              Nop();    // 200 ns
              Nop();    // 250 ns
              Nop();    // 300 ns
              Nop();    // 350 ns
              Nop();    // 400 ns [2.5 MHz]
#endif
#if CPU_MIPS == 10
#define RCK_SPEED_SET
              Nop();    // 100 ns
              Nop();    // 200 ns
              Nop();    // 300 ns
              Nop();    // 400 ns [2.5 MHz]
#endif
#ifndef RCK_SPEED_SET
#error "CPU speed is not supported by this code"
#endif
            RCK_OFF;
    
            // possibly initiate [1] digital pot SPI write
            //
            //   (SPI must be available every 1 msec for display use,
            //    so digital pot SPI writes are done at 1/msec max rate
            //    to guarantee SPI availability)
            //
            // one [word] SPI write to digital pot should take < 10 usec [@ 2.5 MHz]

            if (autosaveBits.L_VOLATILE != 0)
            {
                CS_DIGPOT_ON;          // min 60 nsec to SPI clock start (3 CPU cycles @ 40 MIPS)
                autosaveBits.L_VOLATILE = 0;
                SPI1STATbits.SPIEN = 0;     // must turn off SPI module to change 8/16 mode
                SPI1CON1bits.MODE16 = 1;    // all digital pot writes are 16 bit
                SPI1STATbits.SPIEN = 1;
                SPI1BUF = POT_CMD_WRITE | POT_WIPER_L | settings_gain;
                SPI_mode = 1;
                return; // keep SPI interrupt enabled
            }
            else if (autosaveBits.R_VOLATILE != 0)
            {
                CS_DIGPOT_ON;          // min 60 nsec to SPI clock start (3 CPU cycles @ 40 MIPS)
                autosaveBits.R_VOLATILE = 0;
                SPI1STATbits.SPIEN = 0;     // must turn off SPI module to change 8/16 mode
                SPI1CON1bits.MODE16 = 1;    // all digital pot writes are 16 bit
                SPI1STATbits.SPIEN = 1;
                SPI1BUF = POT_CMD_WRITE | POT_WIPER_R | settings_gain;
                SPI_mode = 1;
                if (settings_gain == 0) CLR_ON  else CLR_OFF
                return; // keep SPI interrupt enabled
            }

            if (autosave_lockout == 0)     // must wait for previous EE write to complete
            {
                if (autosaveBits.L_NONVOLATILE != 0)
                {
                    CS_DIGPOT_ON;          // min 60 nsec to SPI clock start (3 CPU cycles @ 40 MIPS)
                    autosaveBits.L_NONVOLATILE = 0;
                    SPI1STATbits.SPIEN = 0;     // must turn off SPI module to change 8/16 mode
                    SPI1CON1bits.MODE16 = 1;    // all digital pot writes are 16 bit
                    SPI1STATbits.SPIEN = 1;
                    autosave_lockout = 2;  // allow at least 10 msec for EE write to complete
                    SPI1BUF = POT_CMD_WRITE | POT_EE_WIPER | settings_gain;
                    SPI_mode = 1;
                    return; // keep SPI interrupt enabled
                }
                else if (autosaveBits.MODE != 0)
                {
                    CS_DIGPOT_ON;          // min 60 nsec to SPI clock start (3 CPU cycles @ 40 MIPS)
                    autosaveBits.MODE = 0;
                    SPI1STATbits.SPIEN = 0;     // must turn off SPI module to change 8/16 mode
                    SPI1CON1bits.MODE16 = 1;    // all digital pot writes are 16 bit
                    SPI1STATbits.SPIEN = 1;
                    autosave_lockout = 2;  // allow at least 10 msec for EE write to complete
                    SPI1BUF = POT_CMD_WRITE | POT_EE_USER | settings_mode;
                    SPI_mode = 1;
                    return; // keep SPI interrupt enabled
                }
            }

            IEC0bits.SPI1IE = 0;    // end (display) SPI interrupt
        }
    }
}
