;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

    .include "sa.inc" ; hardware registers for particular CPU chip being used

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	.section .libdsp, code

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; _FFT_2048: Complex (in-place) 2048 point FFT
;
; Input:
;	w0 = ptr to complex source buffer (2KW buffer containing 1KW signed integers)
;
; Return:
;   1'st half of source buffer contains the FFT result
;   w0 = data saturation indicatior:
;        b0 = 0
;        b1 = 0
;        b2 = 0
;        b3 = set if saturation occurred in stage 3
;        b4 = set if saturation occurred in stage 4
;        b5 = set if saturation occurred in stage 5
;        b6 = set if saturation occurred in stage 6
;        b7 = set if saturation occurred in stage 7
;        b8 = set if saturation occurred in stage 8
;        b9 = set if saturation occurred in stage 9
;        b10= set if saturation occurred in stage 10
;        b11= set if saturation occurred in stage 11
;
; System resources usage:
;	{w0..w7}	used, not restored
;	{w8..w13}	saved, used, restored
;	 AccuA		used, not restored
;	 AccuB		used, not restored
;	 CORCON		saved, used, restored
;	 PSVPAG		saved, used, restored
;
; DO and REPEAT instruction usage.
;	2 level DO intruction
;	no REPEAT intructions
;
; For an explanation of the algorithm used here, see:
;    Chapter 12- The Fast Fourier Transform
;   "The Scientist and Engineer's Guide to Digital Signal Processing"
;    by Steven W. Smith
;............................................................................


 .global _FFT_2048

_FFT_2048:

;............................................................................

	push.d	w8				; {w8,w9} to TOS
	push.d	w10				; {w10,w11} to TOS
	push.d	w12				; {w12,w13} to TOS

;............................................................................

	; Prepare core for fractional computation
    ; + enable saturation on both accumulators
    ; + enable data space write saturation
    ; + normal saturation (1:31)
    ; + unbiased rounding
    ; + fractional multiply mode
    ; + enable PSV for twiddle factors access
    push	CORCON
    mov     #((1<<SATA)|(1<<SATB)|(1<<SATDW)|(1<<PSV)), w7
    mov     w7, CORCON
    ; Prepare to access twiddle factors
    push    PSVPAG
    mov	    psvpage(#_twiddleAray), w7
    mov	    w7, PSVPAG

;............................................................................

    clr     w1              ; keep saturation indicators on stack
    push    w1

;............................................................................

;-- FFT stage 1
;
;   This is an optimizaton of the generalized FFT algorithm (stages 3-n)
;   realizing that the twiddle factors at the butterflies in this stage are -1,1

    mov     w0,w8           ; working copy of data pointer

    ; (FFT_BLOCK_LENGTH / 2) / 2 - 1  (half-buffer, 2 elements per loop)
    do      #511,_stage1_end

    mov.d   [w8++], w10     ; w10=ar, w11=ai
    mov.d   [w8--], w12     ; w12=br, w13=bi

    ; FFT first stage core calculation

    add     w10,w12, [w8++] ; ar + br
    add     w11,w13, [w8++] ; ai + bi
    sub     w10,w12, [w8++] ; ar - br
_stage1_end: ; loop execution ends at the FOLLOWING INSTRUCTION...
    sub     w11,w13, [w8++] ; ai - bi


;............................................................................

;-- FFT stage 2
;
;   This is an optimizaton of the generalized FFT algorithm (stages 3-n)
;   realizing that the twiddle factors at the butterflies in this stage are -1,1

    mov     w0,w8           ; source pointer
    mov     w0,w9           ; destination pointer

    ; (FFT_BLOCK_LENGTH / 2) / 4 - 1  (half-buffer, 4 elements per loop)
    do      #255,_stage2_end  ; loop does w3+1 iterations

    mov.d   [w8++], w10     ; w10=ar0, w11=ai0
    mov.d   [w8++], w4      ; w4 =ar1, w5 =ai1

    mov.d   [w8++], w12     ; w12=br0, w13=bi0
    mov.d   [w8++], w6      ; w6 =br1, w7 =bi1

    ; FFT second stage core calculation

    add     w10,w12, [w9++] ; ar0 + br0
    add     w11,w13, [w9++] ; ai0 + bi0
    add     w4, w7,  [w9++] ; ar1 + bi1
    sub     w5, w6,  [w9++] ; ai1 - br1

    sub     w10,w12, [w9++] ; ar0 - br0
    sub     w11,w13, [w9++] ; ai0 - bi0
    sub     w4, w7,  [w9++] ; ar1 - bi1
_stage2_end: ; loop execution ends at the FOLLOWING INSTRUCTION...
    add     w5, w6,  [w9++] ; ai1 + br1

;............................................................................

;-- FFT stages 3..10

    .equ    DATAPTR, w0
    .equ    LE2X4,   w1
    .equ    TWIDINC, w2
    .equ    TWIDPTR, w3
    ;       UR     = w4
    ;       UI     = w5
    ;       REX    = w6
    ;       IMX    = w7
    ;       I      = w8
    ;       IP     = w9
    ;       temp   = w10
    ;       temp   = w11
    .equ    LEX4,    w12
    ;       J      = w13

.macro FFT_STAGE LE, TWIDWID, SCALE, SATURATION_BIT
    ;  LE      = number of butterflies in this stage
    ;  TWIDWID = twiddle factor span for each butterfly
    ;  SCALE   = 1 to scale results down by 1/2 (to avoid saturation)
    ;  SAT_BIT = saturation indicator bit number

    ; fast constants
    mov     #(\LE*4),  LEX4  ; LE,   *4 bytes per REX/IMX element
    mov     #(\LE*2),  LE2X4 ; LE/2, *4 bytes per REX/IMX element
    mov     #(\TWIDWID*4), TWIDINC

    ; detect saturation during each phase
    bclr    SR, #SAB        ; clear saturation indicators for accumulators A & B

; FOR J = 0 TO LE2-1
    mov     #_twiddleAray, TWIDPTR
    mov     DATAPTR, w13    ; point to start of REX/IMX array
    do      #(\LE/2-1), _NEXT_J\@

    mov.d   [TWIDPTR], w4   ; w4 = UR, w5 = UI

; FOR I = J TO NM1 STEP LE
    mov     w13, w8         ; I = J
.if (\TWIDWID/2 > 1)
    do      #(\TWIDWID/2-1), _NEXT_I\@
.endif

; FFT core calculation...

    add     LE2X4, w8, w9   ; IP = I + LE2  (LE2 = LE/2 * 4)

    mov.d   [w9], w6        ; w6 = REX[IP], w7 = IMX[IP]

    mpy.n   w4*w6, a        ; a(TR) = -(UR * REX[IP])
    msc     w5*w7, a        ; a(TR) = -(UR * REX[IP]) - (UI * IMX[IP])

.if (\SCALE == 0)
    sac.r   a, w10          ; round TR
    mov     [w8], w11       ; w10=TR, w11=REX[I]
    sub     w11,w10,[w9++]  ; REX[IP] = REX[I] - TR
    add     w11,w10,[w8++]  ; REX[I]  = REX[I] + TR
.else
    lac     [w8], b         ; b = REX[I]
    add     b               ; b = REX[I] + TR
    sac.r   b, #1, w10      ; REX[I] = REX[I] + TR, scaled by 1/2
    lac     [w8], b         ; b = REX[I]
    sub     b               ; b = REX[I] - TR
    sac.r   b, #1, [w9++]   ; REX[IP] = REX[I] - TR, scaled by 1/2
    mov     w10,   [w8++]
.endif

    mpy     w5*w6, a        ; a(TI) =  (UI * REX[IP])
    msc     w4*w7, a        ; a(TI) =  (UI * REX[IP]) - (UR * IMX[IP])

.if (\SCALE == 0)
    sac.r   a, w10          ; round TI
    mov     [w8], w11       ; w10=TI, w11=IMX[I]
    sub     w11,w10,[w9--]  ; IMX[IP] = IMX[I] - TI
    add     w11,w10,[w8--]  ; IMX[I]  = IMX[I] + TI
.else
    lac     [w8], b         ; b = IMX[I]
    add     b               ; b = IMX[I] + TI
    sac.r   b, #1, w10      ; IMX[I] = IMX[I] + TI, scaled by 1/2
    lac     [w8], b         ; b = IMX[I]
    sub     b               ; b = IMX[I] - TI
    sac.r   b, #1, [w9--]   ; IMX[IP] = IMX[I] - TI, scaled by 1/2
    mov     w10,   [w8--]
.endif

_NEXT_I\@: ; loop execution ends at the FOLLOWING INSTRUCTION...
    add     LEX4, w8, w8

    add     TWIDINC, TWIDPTR,TWIDPTR

_NEXT_J\@: ; loop execution ends at the FOLLOWING INSTRUCTION...
    add     #4, w13         ; 4 bytes per REX/IMX pair

    ; IF (saturation) THEN set saturation_bit in return value
    pop     w1              ; recover return value
    btsc    SR, #SAB
    bset    w1, #\SATURATION_BIT
    push    w1              ; store return value

.endm

;-- FFT stage control table for stages 3..10

    ;          LE TWIDWID SCALE SATURATION_BIT
    ;          -- ------- ----- --------------
    FFT_STAGE    8, 256,   0,         3       ; stage 3
    FFT_STAGE   16, 128,   0,         4       ; stage 4
    FFT_STAGE   32,  64,   0,         5       ; stage 5
    FFT_STAGE   64,  32,   0,         6       ; stage 6
    FFT_STAGE  128,  16,   1,         7       ; stage 7
    FFT_STAGE  256,   8,   1,         8       ; stage 8
    FFT_STAGE  512,   4,   1,         9       ; stage 9
    FFT_STAGE 1024,   2,   0,        10       ; stage 10

;............................................................................

;-- even/odd frequency decomposition
;
;   in this stage, the FFT data buffer is expanded from 1KW to the full 2KW

    .equ    IP2,     w2
    .equ    IPM,     w3
    ;       I      = w8
    ;       IM     = w9

    ; switch to super-saturation mode for this phase
    ;  (final result value saturation is not possible in this phase,
    ;   however intermediate calculations may exceed 1:31 math)
    bset    CORCON, #ACCSAT

    ; set up data access indexes
    mov     #(1*4),         w8  ; I   = 1  (4 bytes per REX/IMX element)
    mov     #((2048/2-1)*4),w9  ; IM  = N/2 - I
    mov     #((2048/2+1)*4),IP2 ; IP2 = N/2 + I
    mov     #((2048-1)*4),  IPM ; IPM = N/2 + IM
    ; make 'em all pointers to the REX/IMX data array
    add     w0, w8,w8
    add     w0, w9,w9
    add     w0, IP2,IP2
    add     w0, IPM,IPM

    ; FOR I = 1 TO N/4-1
    do      #((2048/4-1)-1), _NEXT_I

    mov.d   [w8], w4        ; w4 = REX[I],  w5 = IMX[I]
    mov.d   [w9], w6        ; w6 = REX[IM], w7 = IMX[IM]

    ; REX[IP2] = (IMX[I] + IMX[IM]) / 2
    lac     w5, a
    add     w7, a
    sac.r   a, #1, [IP2++]
    ; REX[IPM] = REX[IP2]
    sac.r   a, #1, [IPM++]

    ; IMX[IP2] = -(REX[I] - REX[IM]) / 2 == (REX[IM] - REX[I]) / 2
    lac     w4, a
    neg     a
    add     w6, a
    sac.r   a, #1, [IP2--]
    ; IMX[IPM] = -IMX[IP2] == (REX[I] - REX[IM]) / 2
    neg     a
    sac.r   a, #1, [IPM--]

    ; REX[I] = (REX[I] + REX[IM]) / 2
    lac     w4, a
    add     w6, a
    sac.r   a, #1, [w8++]
    ; REX[IM] = REX[I]
    sac.r   a, #1, [w9++]

    ; IMX[I] = (IMX[I] - IMX[IM]) / 2
    lac     w7, a
    neg     a
    add     w5, a
    sac.r   a, #1, [w8--]
    ; IMX[IM] = -IMX[I] == (IMX[IM] - IMX[I]) / 2
    neg     a
    sac.r   a, #1, [w9--]

    sub     #4, w9
    add     #4, IP2
    sub     #4, IPM
_NEXT_I: ; loop execution ends at the FOLLOWING INSTRUCTION...
    add     #4, w8          ; I = I+1  (4 bytes per REX/IMX element)

    ; REX[N*3/4] = IMX[N/4]
    mov     #((2048/4)*4+2), w2
    add     w0, w2, w2
    mov     #((2048*3/4)*4), w1
    add     w0, w1, w1
    mov     [w2], [w1]

    ; REX[N/2] = IMX[0]
    add     w0, #2, w2
    mov     #((2048/2)*4), w1
    add     w0, w1, w1
    mov     [w2], [w1]

    ; IMX[N*3/4] = 0
    mov     #((2048*3/4)*4+2), w1
    add     w0, w1, w1
    clr     [w1]

    ; IMX[N/2] = 0
    mov     #((2048/2)*4+2), w1
    add     w0, w1, w1
    clr     [w1]

    ; IMX[N/4] = 0
    mov     #((2048/4)*4+2), w1
    add     w0, w1, w1
    clr     [w1]

    ; IMX[0] = 0
    add     w0, #2, w1
    clr     [w1]

    ; restore normal saturation mode (1:31 math)
    bclr    CORCON, #ACCSAT

;............................................................................

;-- FFT stage 11
;
;   operating on the entire (2KW) FFT data buffer in this stage

    .equ    SCALE_STAGE11, 0
    .equ    LE_STAGE11, 2048

    ; fast constants
    mov     #(LE_STAGE11*2), LE2X4 ; LE/2, *4 bytes per REX/IMX element

    ; detect saturation during each phase
    bclr    SR, #SAB        ; clear saturation indicators for accumulators A & B

    ; FOR J = 0 TO LE2-1
    mov     #_twiddleAray, TWIDPTR
    mov     w0, w8          ; point to start of REX/IMX array

    do      #(LE_STAGE11/2-1), _NEXT_J_STAGE11

    mov.d   [TWIDPTR++], w4 ; w4 = UR, w5 = UI

; FFT core calculation...

    add     LE2X4, w8, w9   ; IP = I + LE2  (LE2 = LE/2 * 4)

    mov.d   [w9], w6        ; w6 = REX[IP], w7 = IMX[IP]

    mpy.n   w4*w6, a        ; a(TR) = -(UR * REX[IP])
    msc     w5*w7, a        ; a(TR) = -(UR * REX[IP]) - (UI * IMX[IP])

.if (SCALE_STAGE11 == 0)
    sac.r   a, w10          ; round TR
    mov     [w8], w11       ; w10=TR, w11=REX[I]
    add     w11,w10,[w8++]  ; REX[I]  = REX[I] + TR
.else
    lac     [w8], b         ; b = REX[I]
    add     b               ; b = REX[I] + TR
    sac.r   b, #1,[w8++]    ; REX[I] = REX[I] + TR, scaled by 1/2
.endif

    mpy     w5*w6, a        ; a(TI) =  (UI * REX[IP])
    msc     w4*w7, a        ; a(TI) =  (UI * REX[IP]) - (UR * IMX[IP])

.if (SCALE_STAGE11 == 0)
    sac.r   a, w10          ; round TI
    mov     [w8], w11       ; w10=TI, w11=IMX[I]
    add     w11,w10,[w8--]  ; IMX[I]  = IMX[I] + TI
.else
    lac     [w8], b         ; b = IMX[I]
    add     b               ; b = IMX[I] + TI
    sac.r   b, #1, [w8--]   ; IMX[I] = IMX[I] + TI, scaled by 1/2
.endif

_NEXT_J_STAGE11: ; loop execution ends at the FOLLOWING INSTRUCTION...
    add     #4, w8          ; 4 bytes per REX/IMX pair

    ; IF (saturation) THEN set saturation_bit in return value
    pop     w1              ; recover return value
    btsc    SR, #SAB
    bset    w1, #11
    push    w1              ; store return value

;............................................................................

    ; get return value
    pop     w0              ; return saturation indicators

	; restore core settings
	pop	PSVPAG
	pop	CORCON

;............................................................................

	; restore working registers
	pop.d	w12				; {w12,w13} from TOS
	pop.d	w10				; {w10,w11} from TOS
	pop.d	w8				; {w8,w9} from TOS

;............................................................................

	return	

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	.end

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
